import numpy as np
from scipy.interpolate import interp1d


def resample_to_uniform_offset(offset_in, value_in, offset_out):

    # size of input
    nrow, ncol = offset_in.shape

    # initiate output
    value_out = np.ones((nrow, len(offset_out))) * 0

    # check input type
    if np.iscomplex(value_in).any():
        value_out = value_out.astype(complex)

    # loop across rows
    for ii in np.arange(nrow):

        # interpolate value
        xx = offset_in[ii, :]
        yy = value_in[ii, :]

        # ignore nan
        kk = ~np.isnan(xx * yy)
        xx = xx[kk]
        yy = yy[kk]

        # ensure strictly input offset
        kk = np.where(np.diff(xx) > 0)[0]
        if len(kk) >= 2:

            # strictly increasing part
            xx = xx[kk]
            yy = yy[kk]

            # interpolate
            ff = interp1d(xx, yy, kind="cubic", bounds_error=False, fill_value=np.nan)
            val = ff(offset_out)
            value_out[ii, :] = val

    return value_out



import numpy as np
from scipy import signal
from collections import namedtuple


def get_wavelet(
    wav_duration=0.2,
    wav_dt=0.004,
    wav_type="ricker",
    wav_parameters={"wav_ricker_freq": 30},
):

    # - initiate output
    wav_time = np.nan
    wav_amplitude = np.nan

    # get wavelet based on type
    if wav_type == "ricker":

        # - assign input parameters
        wav_ricker_freq = wav_parameters["wav_ricker_freq"]

        # - get wavelet in time domain
        ww = _ricker(wav_duration, wav_dt, wav_ricker_freq)
        wav_time = ww.time  # wavelet time
        wav_amplitude = ww.amplitude  # wavelet amplitude

        # - make string
        wav_name = "Ricker (" + str(wav_ricker_freq) + " Hz)"

    elif wav_type == "ormsby":

        # - assign input parameters
        wav_ormsby_freq = wav_parameters["wav_ormsby_freq"]

        # - get wavelet in time domain
        ww = _ormsby(wav_duration, wav_dt, wav_ormsby_freq)
        wav_time = ww.time  # wavelet time
        wav_amplitude = ww.amplitude  # wavelet amplitude

        # - make string
        ss = "".join([str(ff) + "-" for ff in wav_ormsby_freq])
        wav_name = "Ormsby (" + ss[:-1] + " Hz)"

        # - apply hanning filter
        hann_filter = signal.windows.hann(len(wav_time) + 1)
        wav_amplitude = wav_amplitude * hann_filter[:-1]

    # return wavelet in time domain
    return wav_time, wav_amplitude, wav_name


def time_to_frequency(wav_time, wav_amplitude, crop_positive=False):

    # Obtain frequency power spectrum from wavelet in time

    # - time vector
    tv = wav_time
    dt = tv[1] - tv[0]
    ntv = len(tv)

    # - define frequency vector for FFT
    nf = int(pow(2, np.ceil(np.log(ntv) / np.log(2))))
    df = 1 / (dt * nf)
    # frequency sampling
    fmin = -0.5 * df * nf
    fv = fmin + df * np.arange(0, nf)  # frequency vector

    # - wavelet in f-domain
    ww_ap = np.zeros(nf)  # initilize wavelet amplitude padded with zeros
    ww_ap[: len(wav_amplitude)] = wav_amplitude  # insert wavelet amplitude
    ww_af = np.fft.fft(ww_ap)  # wavelet in frequency domain
    ww_af = np.fft.fftshift(ww_af)

    # - crop freq vector and normalize power spectrum
    if crop_positive:
        ww_af = ww_af[fv >= 0]
        wav_frequency = fv[fv >= 0]
        wav_spectrum = np.abs(ww_af) / max(np.abs(ww_af))

    # - return
    return wav_frequency, wav_spectrum


def rotate_phase(w, phi, degrees=False):
    r"""
    Performs a phase rotation of wavelet or wavelet bank using:

    .. math::

        A = w(t)\cos\phi - h(t)\sin\phi

    where `w(t)` is the wavelet, `h(t)` is its Hilbert transform, and \phi is
    the phase rotation angle (default is radians).

    The analytic signal can be written in the form :math:`S(t) = A(t)e^{j\theta (t)}`
    where :math:`A(t) = \left| h(w(t)) \right|` and :math:`\theta(t) = \tan^{-1}[h(w(t))]`. 
    `A(t)` is called the "reflection strength" and :math:`\phi(t)` is called the "instantaneous
    phase".

    A constant phase rotation :math:`\phi` would produce the analytic signal
    :math:`S(t)=A(t)e^{j(\theta(t) + \phi)}`. To get the non-analytic signal,
    we take 

    .. math::

        real(S(t)) &= A(t)\cos(\theta(t) + \phi) \\
        &= A(t)\cos\theta(t)\cos(\phi)-\sin\theta(t)\sin(\phi))\\
        &= w(t)\cos\phi-h(t)\sin\phi
        

    Args:
        w (ndarray): The wavelet vector, can be a 2D wavelet bank.
        phi (float): The phase rotation angle (in radians) to apply.
        degrees (bool): If phi is in degrees not radians.

    Returns:
        The phase rotated signal (or bank of signals).
    """
    # Make sure the data is at least 2D to apply_along
    data = np.atleast_2d(w)

    # Get Hilbert transform. This will be 2D.
    a = _apply_along_axis(signal.hilbert, data, axis=0)

    # Transform angles into what we need.
    phi = np.asanyarray(phi).reshape(-1, 1, 1)
    if degrees:
        phi = np.radians(phi)

    rotated = np.real(a) * np.cos(phi) - np.imag(a) * np.sin(phi)
    return np.squeeze(rotated)


def _apply_along_axis(func_1d, arr, *args, **kwargs):
    """
    Apply 1D function across 2D slice as efficiently as possible.

    Although `np.apply_along_axis` seems to do well enough, map usually
    seems to end up beig a bit faster.

    Args:
        func_1d (function): the 1D function to apply, e.g. np.convolve. Should
            take 2 or more arguments: the

    Example
    >>> apply_along_axes(np.convolve, reflectivity_2d, wavelet, mode='same')
    """
    mapobj = map(lambda tr: func_1d(tr, *args, **kwargs), arr)
    return np.array(list(mapobj))


def _get_time(duration, dt):
    """
    Make a time vector.

    The time vector will have an odd number of samples,
    and will be symmetric about 0.
    """
    # This business is to avoid some of the issues with `np.arange`:
    # (1) unpredictable length and (2) floating point weirdness, like
    # 1.234e-17 instead of 0. Not using `linspace` because figuring out
    # the length and offset gave me even more of a headache than this.
    n = int(duration / dt)
    odd = n % 2
    k = int(10 ** -np.floor(np.log10(dt)))
    dti = int(k * dt)  # integer dt

    if odd:
        t = np.arange(n)
    elif not odd:
        t = np.arange(n + 1)

    t -= t[-1] // 2

    return dti * t / k


def _ricker(duration, dt, f):
    """
    Also known as the mexican hat wavelet, models the function:

    .. math::
        A =  (1 - 2 \pi^2 f^2 t^2) e^{-\pi^2 f^2 t^2}

    .. plot::

        import matplotlib.pyplot as plt
        w, t = _ricker(0.256, 0.002, 40)
        plt.plot(t, w)

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (often one of  0.001, 0.002,
            or 0.004).
        f (float): Centre frequency of the wavelet in Hz.

    Returns:
        ndarray: A vector containing a tuple of (wavelet, t) is returned.

    """
    t = _get_time(duration, dt)

    pft2 = (np.pi * f * t) ** 2
    w = np.squeeze((1 - (2 * pft2)) * np.exp(-pft2))

    RickerWavelet = namedtuple("RickerWavelet", ["amplitude", "time"])
    return RickerWavelet(w, t)


def _ormsby(duration, dt, f):
    """
    The Ormsby wavelet requires four frequencies which together define a
    trapezoid shape in the spectrum. The Ormsby wavelet has several sidelobes,
    unlike Ricker wavelets.

    .. plot::

        import matplotlib.pyplot as plt
        w, t = _ormsby(0.256, 0.002, [5, 10, 40, 80])
        plt.plot(t, w)

    Args:
        duration (float): The length in seconds of the wavelet.
        dt (float): The sample interval in seconds (usually 0.001, 0.002,
            or 0.004).
        f (array-like): A tuple of frequencies (f1, f2, f3, f4).
    Returns:
        ndarray: A vector containing a tuple of (wavelet, t) is returned.

    """
    try:
        f = np.array(f).reshape(4)
    except ValueError:
        raise ValueError("The dimension of the frequency array must be of size 4.")

    t = _get_time(duration, dt)
    f1, f2, f3, f4 = f

    def numerator(f, t):
        """The numerator in the Ormsby equation."""
        return (np.sinc(f * t) ** 2) * ((np.pi * f) ** 2)

    pf43 = (np.pi * f4) - (np.pi * f3)
    pf21 = (np.pi * f2) - (np.pi * f1)

    w = (
        (numerator(f4, t) / pf43)
        - (numerator(f3, t) / pf43)
        - (numerator(f2, t) / pf21)
        + (numerator(f1, t) / pf21)
    )

    OrmsbyWavelet = namedtuple("OrmsbyWavelet", ["amplitude", "time"])
    return OrmsbyWavelet(np.squeeze(w) / np.amax(w), t)



import numpy as np


def conv_vti_thoms2cij(rho, vp, vs, delT=[], epsT=[], gamT=[]):
    
    # print(vp[:5])
    # print(delT[:5])

    # if delT is None:
    #     delT = np.zeros(np.shape(rho))
    # if epsT is None:
    #     epsT = np.zeros(np.shape(rho))
    # if gamT is None:
    #     gamT = np.zeros(np.shape(rho))
        
    # include anisotropy
    nn = len(rho)
    if len(delT) == 0:
        delT = np.zeros(nn)
    if len(epsT) == 0:
        epsT = np.zeros(nn)
    if len(gamT) == 0:
        gamT = np.zeros(nn)
        

    c33 = rho * vp ** 2
    c44 = rho * vs ** 2
    c11 = c33 * (2 * epsT + 1)
    c66 = c44 * (2 * gamT + 1)
    c13 = np.sqrt(2 * c33 * (c33 - c44) * delT + (c33 - c44) ** 2) - c44

    return rho, c11, c13, c33, c44, c66


def conv_vti_cij2_thoms(rho, c11, c13, c33, c44, c66):

    vp = np.sqrt(c33 / rho)  # vertical P-wave velocity (m/s)
    vs = np.sqrt(c44 / rho)  # vertical S-wave velocity (m/s)
    delT = ((c13 + c44) ** 2 - (c33 - c44) ** 2) / (2 * c33 * (c33 - c44))
    epsT = (c11 - c33) / (2 * c33)
    gamT = (c66 - c44) / (2 * c44)

    return vp, vs, delT, epsT, gamT



import setuptools
import seismod1d

long_description = """
Create synthetic seismic
"""

setuptools.setup(
    name="seismod1d",
    version=seismod1d.__version__,
    author="Equinor ASA",
    author_email="eidi@equinor.com",
    description=long_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/equinor/seismod1d",
    packages=setuptools.find_packages(),
    package_data={"seismod1d": ["colormaps/*.dat"]},
    include_package_data=True,
    install_requires=[
        "numpy",
        "pandas",
        "matplotlib",
        "scipy",
        "lasio",
        "segyio",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



import numpy as np


def iso(depth, vp, theta_in=np.arange(90)):

    # fast isotropic raytracer

    # convert input angles to radians
    theta_in = np.radians(theta_in)

    # get input size
    nlayers = len(vp)
    nrays = len(theta_in)

    # layer thicknesses
    dz = np.diff(depth)
    dz = np.insert(
        dz, 0, 0
    )  # insert dummy 'top thickness' to make arrays same lengths (not used)

    # get ray angle (phase = group) for all rays and layers
    m = vp.reshape((nlayers, 1)) * np.sin(theta_in) / vp[0]
    m[m > 1] = np.nan  # set post-critical angles to nan
    phase_angle = np.arcsin(m)

    # get offset
    doffset = 2 * dz.reshape((nlayers, 1)) * np.tan(phase_angle)
    offset = np.cumsum(doffset, axis=0)

    # get time
    dt = 2 * dz / vp  # zero offset time step
    dtime = dt.reshape((nlayers, 1)) * (np.cos(phase_angle) ** -1)
    time = np.cumsum(dtime, axis=0)

    # define additional output
    phase_angle = np.degrees(phase_angle)
    phase_velocity = np.tile(vp.reshape((len(vp), 1)), (1, nrays)).astype(float)
    phase_velocity[np.isnan(phase_angle)] = np.nan
    group_angle = np.degrees(phase_angle)
    group_velocity = phase_velocity

    return offset, time, phase_angle, group_angle, phase_velocity, group_velocity


def vti(depth, vp, vs, delta, epsilon, theta_in=np.arange(90)):

    # 1D VTI raytracer based on Tang & Li (2008) and Tsvankin(1996)

    # turn off some warnings
    np.seterr(divide="ignore", invalid="ignore")

    # convert input angles to radians
    theta_in = np.radians(theta_in)

    # get input size
    nlayers = len(vp)
    nrays = len(theta_in)

    # set anisotropy arrays to zero if empty
    if len(delta) == 0:
        delta = 0 * vp
    if len(epsilon) == 0:
        epsilon = 0 * vp

    # initiate output
    dtype = "complex"
    offset = np.zeros((nlayers, nrays)).astype("float")  # m
    time = np.zeros((nlayers, nrays)).astype("float")  # s
    phase_angle = np.zeros((nlayers, nrays)).astype(dtype)  # radians
    group_angle = np.zeros((nlayers, nrays)).astype(dtype)  # radians
    phase_velocity = np.zeros((nlayers, nrays)).astype(dtype)  # m/s
    group_velocity = np.zeros((nlayers, nrays)).astype(dtype)  # m/s

    # layer thicknesses
    dz = np.diff(depth)
    dz = np.insert(
        dz, 0, 0
    )  # insert dummy 'top thickness' to make arrays same lengths (not used)

    # get initial slowness
    pp = np.sin(theta_in) / get_phase_velocity(
        theta_in, vp[1], vs[1], delta[1], epsilon[1]
    )

    # loop through layers
    for ii in np.arange(nlayers):

        # get phase angles (from slowness)new_model
        th_ph = get_phase_angle(pp, vp[ii], vs[ii], delta[ii], epsilon[ii])

        # get phase velocity
        vp_ph = get_phase_velocity(th_ph, vp[ii], vs[ii], delta[ii], epsilon[ii])

        # set post-critical angles to NaN
        if 1:
            kk = np.imag(th_ph) != 0
            th_ph[kk] = np.nan
            pp[kk] = np.nan
            vp_ph[kk] = np.nan

        # get group velocity
        vp_gr = get_group_velocity(th_ph, vp_ph, vp[ii], vs[ii], delta[ii], epsilon[ii])

        # get group angle
        th_gr = get_group_angle(th_ph, vp_ph, vp[ii], vs[ii], delta[ii], epsilon[ii])

        # claculate traveltime and offset increments
        dt = 2 * dz[ii] / (vp_gr * np.cos(th_gr))
        dx = 2 * dz[ii] * np.tan(th_gr)

        # update output
        time[ii, :] = time[ii - 1, :] + np.real(dt)
        offset[ii, :] = offset[ii - 1, :] + np.real(dx)
        phase_angle[ii, :] = th_ph * (180 / np.pi)  # rad --> degree
        group_angle[ii, :] = th_gr * (180 / np.pi)  # rad --> degree
        phase_velocity[ii, :] = vp_ph
        group_velocity[ii, :] = vp_gr

    # turn back on warnings
    np.seterr(all="warn")

    return offset, time, phase_angle, group_angle, phase_velocity, group_velocity


def get_phase_angle(p, vp0, vs0, delT, epsT):

    # initiate output
    theta_phase = np.ones(len(p)).astype("complex") * np.nan

    # calculate phase angle
    f = 1 - (vs0 / vp0) ** 2
    c0 = (1 - f) * vp0 ** 4 * p ** 4
    c1 = (2 * (epsT - f * delT) * vp0 ** 2 * p ** 2 - 2 + f) * vp0 ** 2 * p ** 2
    c2 = 1 - 2 * epsT * vp0 ** 2 * p ** 2 - 2 * f * (epsT - delT) * vp0 ** 4 * p ** 4
    temp = (-c1 + np.sqrt(c1 ** 2 - 4 * c0 * c2)) / (2 * c2)
    temp_sq = np.sqrt(temp)

    # update phase angle
    kk = temp_sq <= 1  # identiy critical angles
    theta_phase[kk] = np.arcsin(temp_sq[kk])

    # theta_phase = np.arcsin(temp_sq)

    return theta_phase


def get_phase_velocity(th, vp0, vs0, delT, epsT):

    f = 1 - (vs0 / vp0) ** 2
    temp = (
        1
        + (4 / f)
        * np.sin(th) ** 2
        * (2 * delT * np.cos(th) ** 2 - epsT * np.cos(2 * th))
        + 4 * (epsT / f) ** 2 * np.sin(th) ** 4
    )
    temp = 1 + epsT * np.sin(th) ** 2 - (f / 2) + (f / 2) * np.sqrt(temp)
    vp_phase = vp0 * np.sqrt(temp)

    return vp_phase


def get_phase_vel_sq_deriv(th, vp0, vs0, delT, epsT):

    # derivate of velocity squared in terms of phase angle
    f = 1 - (vs0 / vp0) ** 2
    d1 = 1 + (2 * epsT * np.sin(th) ** 2) / f
    d2 = 2 * (epsT - delT) * np.sin(2 * th) ** 2 / f
    D = np.sqrt(d1 ** 2 - d2)
    E = epsT * (f + 2 * epsT * np.sin(th) ** 2) * np.sin(2 * th) - f * (
        epsT - delT
    ) * np.sin(4 * th)
    Dvp_ph_sq = vp0 ** 2 * (epsT * np.sin(2 * th) + E / (f * D))

    return Dvp_ph_sq


def get_group_velocity(th, vp_ph, vp0, vs0, delT, epsT):

    # get derivative of square of phase velocity
    Dvp_ph_sq = get_phase_vel_sq_deriv(th, vp0, vs0, delT, epsT)

    # group velocity
    vp_gr = np.sqrt(vp_ph ** 2 + (4 * vp_ph ** 2) ** (-1) * (Dvp_ph_sq ** 2))

    return vp_gr


def get_group_angle(th, vp_ph, vp0, vs0, delT, epsT):

    # get derivative of square of phase velocity
    Dvp_ph_sq = get_phase_vel_sq_deriv(th, vp0, vs0, delT, epsT)

    # group velocity
    dth = np.arctan((2 * vp_ph ** 2) ** (-1) * Dvp_ph_sq)
    th_gr = th + dth

    return th_gr



import numpy as np


def generate_test_model(model_version=1):
    """Generates a test model for synthmod 1d

    Parameters
    ----------
    model_version : int, optional
        Model version, by default 1

    Returns
    -------
    Tuple[np.ndarray]
        depth, vp, vs, rhob, delta, epsilon, gamma
    """
    if model_version == 1:

        vp = [1480, 3000, 3200, 3500]
        vs = [0, 1500, 1900, 2000]
        rhob = [1, 2.55, 2.2, 2.6]
        delta = [0, 0.1, 0, 0.08]
        thickness = [300, 1000, 200, 1000]

    elif model_version == 2:
        # hard layer creating critical angle
        vp = [1480, 3000, 5500, 3200, 3500]
        vs = [0, 1500, 3500, 1900, 2000]
        rhob = [1, 2.55, 2.9, 2.2, 2.6]
        delta = [0, 0.1, 0, 0, 0.08]
        thickness = [300, 1000, 200, 200, 1000]

    epsilon = list(2 * np.array(delta))
    gamma = list(3 * np.array(delta))
    depth = np.cumsum(thickness)

    # double first layer (values defined as down-to)
    vp = np.insert(vp, 0, vp[0])
    vs = np.insert(vs, 0, vs[0])
    rhob = np.insert(rhob, 0, rhob[0])
    delta = np.insert(delta, 0, delta[0])
    epsilon = np.insert(epsilon, 0, epsilon[0])
    gamma = np.insert(gamma, 0, gamma[0])
    depth = np.insert(depth, 0, 0)

    return depth, vp, vs, rhob, delta, epsilon, gamma
